@@ -17,9 +17,8 @@
 """CNN2SNN objects required to serialize Keras quantized models.
 """
 
-from .quantization_ops import (StdWeightQuantizer, TrainableStdWeightQuantizer,
-                               MaxQuantizer, MaxPerAxisQuantizer,
-                               ScaleFactorQuantizer)
+from .quantization_ops import (StdWeightQuantizer, StdPerAxisQuantizer,
+                               MaxQuantizer, MaxPerAxisQuantizer)
 from .quantization_layers import (QuantizedConv2D, QuantizedSeparableConv2D,
                                   QuantizedDense, ActivationDiscreteRelu,
                                   QuantizedReLU)
@@ -27,10 +26,9 @@ from .min_value_constraint import MinValueConstraint
 
 cnn2snn_objects = {
     'StdWeightQuantizer': StdWeightQuantizer,
-    'TrainableStdWeightQuantizer': TrainableStdWeightQuantizer,
+    'StdPerAxisQuantizer': StdPerAxisQuantizer,
     'MaxQuantizer': MaxQuantizer,
     'MaxPerAxisQuantizer': MaxPerAxisQuantizer,
-    'ScaleFactorQuantizer': ScaleFactorQuantizer,
     'QuantizedConv2D': QuantizedConv2D,
     'QuantizedSeparableConv2D': QuantizedSeparableConv2D,
     'QuantizedDense': QuantizedDense,
