@@ -15,12 +15,12 @@
 # limitations under the License.
 # ******************************************************************************
 """
-Custom tf.keras.constraint.Constraint object to limit weights minimum value.
+Custom keras.constraint.Constraint object to limit weights minimum value.
 """
 
 import tensorflow as tf
 
-from keras.constraints import Constraint
+from tf_keras.constraints import Constraint
 
 
 class MinValueConstraint(Constraint):
@@ -38,4 +38,6 @@ class MinValueConstraint(Constraint):
         return tf.clip_by_value(w, self.min_value, tf.float32.max)
 
     def get_config(self):
-        return {'min_value': self.min_value}
+        config = super().get_config()
+        config.update({'min_value': self.min_value})
+        return config
