@@ -30,7 +30,7 @@ from .quantization_ops import ceil_through
 def _check_unsupported_args(kwargs, unsupported_args):
     """Raises error if unsupported argument are present in kwargs.
 
-    Unsupported arguments: 'data_format', 'activation', 'depth_mutiplier'.
+    Unsupported arguments: 'data_format', 'activation', 'depth_multiplier'.
 
     Args:
         kwargs (dictionary): keyword arguments to check.
@@ -338,7 +338,7 @@ class QuantizedDense(layers.Dense):
             # operation for large sparse input tensors. The op will result in a
             # sparse gradient, as opposed to
             # sparse_ops.sparse_tensor_dense_matmul which results in dense
-            # gradients. This can lead to sigfinicant speedups, see b/171762937.
+            # gradients. This can lead to significant speedups, see b/171762937.
             if isinstance(inputs, tf.SparseTensor):
                 raise TypeError(
                     f"The inputs of a Dense Layer must be a uniform \
@@ -597,7 +597,7 @@ class ActivationDiscreteRelu(QuantizedActivation):
 
         - threshold is 0.5 and step is 1
 
-    For bithwidth > 2, with N = 2^bitwidth - 1:
+    For bitwidth > 2, with N = 2^bitwidth - 1:
 
         - threshold is 3 / N and step is 6 / N
 
